from __future__ import annotations
from typing import Dict, Iterable, Iterator, Optional

from .enumerable import Enumerable
from .types_linq_error import InvalidOperationError

from .more_typing import (
    TSource_co,
)


class CachedEnumerable(Enumerable[TSource_co]):
    '''
    ```py
    from types_linq.cached_enumerable import CachedEnumerable
    ```

    Enumerable that stores the enumerated results which can be accessed repeatedly.

    Users should not construct instances of this class directly. Use `Enumerable.as_cached()` instead.

    Revisions
        ~ v0.1.1: New.
    '''

    _iter: Iterator[TSource_co]
    _cache_capacity: Optional[int]
    _enumerated_values: Dict[int, TSource_co]
    _min_index: int
    _tracked: int

    def __init__(self, source: Iterable[TSource_co], cache_capacity: Optional[int]):
        if cache_capacity is not None and cache_capacity < 0:
            raise InvalidOperationError('cache_capacity must be nonnegative')
        self._iter = iter(source)
        super().__init__(self._iter)
        self._cache_capacity = cache_capacity
        self._enumerated_values = {}
        self._min_index = 0
        self._tracked = 0

    def _get_iterable(self) -> Iterator[TSource_co]:
        i = 0
        while True:
            while i < self._tracked and self._cache_capacity != 0:
                i = max(self._min_index, i)
                res = self._enumerated_values[i]
                i += 1
                yield res
            try:
                res = next(self._iter)
            except StopIteration:
                break
            if self._cache_capacity == 0:
                yield res
                continue
            len_ = len(self._enumerated_values)
            if self._cache_capacity is not None and \
                len_ > 0 and len_ == self._cache_capacity:
                del self._enumerated_values[self._min_index]
                self._min_index += 1
            self._enumerated_values[self._tracked] = res
            self._tracked += 1
            i += 1
            yield res

    def as_cached(self, *, cache_capacity: Optional[int] = None) -> CachedEnumerable[TSource_co]:
        '''
        Updates settings and returns the original CachedEnumerable reference.

        Raises `InvalidOperationError` if cache_capacity is negative.
        '''
        if cache_capacity is not None:
            if cache_capacity < 0:
                raise InvalidOperationError('cache_capacity must be nonnegative')
            while len(self._enumerated_values) > cache_capacity:
                del self._enumerated_values[self._min_index]
                self._min_index += 1
        self._cache_capacity = cache_capacity
        return self



from __future__ import annotations
from typing import Any, Callable, Deque, Generic, Iterable, Optional, Union

from .more_enumerable import MoreEnumerable
from ..more_typing import (
    TKey,
    TSource,
    TSource_co,
)


# appreciation to: MoreLinq.IExtremaEnumerable by morelinq
class ExtremaEnumerable(MoreEnumerable[TSource_co], Generic[TSource_co, TKey]):

    _selector: Callable[[TSource_co], TKey]
    _comparer: Optional[Callable[[TKey, TKey], int]]
    _for_min: bool

    def __init__(self,
        source: Iterable[TSource_co],
        selector: Callable[[TSource_co], TKey],
        comparer: Optional[Callable[[TKey, TKey], int]],
        for_min: bool,
    ) -> None:
        super().__init__(source)
        self._selector = selector
        self._comparer = comparer
        self._for_min = for_min

    # only enumerate extrema when needed. __iter__-ing it will get all extrema; take()-ing it with
    # count elems only store that number of elems in the buffer
    def _get_iterable(self) -> Iterable[TSource_co]:
        it = super()._get_iterable()
        getter = lambda: _FirstExtrema(None)
        return _extrema_by(it, getter, self._selector, self._comparer, self._for_min)

    # please maintain same overloads
    def first(self, *args):  # pyright: ignore[reportIncompatibleMethodOverride]
        if len(args) == 0:
            return MoreEnumerable.first(self.take(1))
        else:  # len(args) == 1
            return super().first(*args)

    # please maintain same overloads
    def first2(self, *args):
        if len(args) == 1:
            return MoreEnumerable.first2(self.take(1), *args)
        else:  # len(args) == 2
            return super().first2(*args)

    # please maintain same overloads
    def last(self, *args):
        if len(args) == 0:
            return MoreEnumerable.last(self.take_last(1))
        else:  # len(args) == 1
            return super().last(*args)

    # please maintain same overloads
    def last2(self, *args):
        if len(args) == 1:
            return MoreEnumerable.last2(self.take_last(1), *args)
        else:  # len(args) == 2
            return super().last2(*args)

    # please maintain same overloads
    def single(self, *args):
        if len(args) == 0:
            return MoreEnumerable.single(self.take(2))
        else:  # len(args) == 1
            return super().single(*args)

    # please maintain same overloads
    def single2(self, *args):
        if len(args) == 1:
            return MoreEnumerable.single2(self.take(2), *args)
        else:  # len(args) == 2
            return super().single2(*args)

    # please maintain same overloads
    def take(self, count: Union[int, slice]) -> Any:
        if isinstance(count, int):
            if count <= 0:
                return MoreEnumerable(())
            it = super()._get_iterable()
            getter = lambda c=count: _FirstExtrema(c)
            return _extrema_by(it, getter, self._selector, self._comparer, self._for_min)
        else:  # isinstance(count, slice)
            return super().take(count)

    def take_last(self, count: int) -> MoreEnumerable[TSource_co]:
        if count <= 0:
            return MoreEnumerable(())
        it = super()._get_iterable()
        getter = lambda: _LastExtrema(count)
        return _extrema_by(it, getter, self._selector, self._comparer, self._for_min)


def _extrema_by(
    source: Iterable[TSource],
    extrema_getter: Callable[[], _Extrema[TSource]],
    selector: Callable[[TSource], TKey],
    comparer: Optional[Callable[[TKey, TKey], int]],
    for_min: bool,
) -> MoreEnumerable[TSource]:
    it = iter(source)
    try:
        elem = next(it)
    except StopIteration:
        return MoreEnumerable(())

    if comparer is None:
        # None comparer ensures TSource must support __lt__()
        comparer = _lt_comparer
    if for_min:
        comparer = lambda x, y, comp=comparer: comp(y, x)

    extrema = extrema_getter()
    extrema.add(elem)
    extremum_key = selector(elem)
    for elem in it:
        key = selector(elem)
        cmp = comparer(key, extremum_key)
        if cmp > 0:
            extrema.restart()
            extrema.add(elem)
            extremum_key = key
        elif cmp == 0:
            extrema.add(elem)

    def inner():
        yield from extrema.store
    return MoreEnumerable(inner)


def _lt_comparer(x, y) -> int:
    if x < y: return -1
    # assumes total ordering
    elif y < x: return 1
    return 0


class _FirstExtrema(Generic[TSource]):
    def __init__(self, capacity: Optional[int]) -> None:
        self.store = []
        self.capacity = capacity

    def add(self, element: TSource) -> None:
        if self.capacity is None or len(self.store) < self.capacity:
            self.store.append(element)

    def restart(self) -> None:
        self.store.clear()


class _LastExtrema(Generic[TSource]):
    def __init__(self, capacity: Optional[int]) -> None:
        self.store = Deque()
        self.capacity = capacity

    def add(self, element: TSource) -> None:
        if self.capacity is not None and len(self.store) == self.capacity:
            self.store.popleft()
        self.store.append(element)

    def restart(self) -> None:
        self.store.clear()


_Extrema = Union[_FirstExtrema[TSource], _LastExtrema[TSource]]



'''
Typing utilities used by methods's declarations across the library. For more details, see
[`typing`](https://docs.python.org/3/library/typing.html).
```{note} Definitions in this module are for documenting purposes only.
```
'''
import sys
from abc import ABCMeta, abstractmethod
from typing import Any, TypeVar

if sys.version_info >= (3, 8, 0):
    from typing import Protocol, runtime_checkable
else:
    from typing_extensions import Protocol, runtime_checkable  # type: ignore


TAccumulate = TypeVar('TAccumulate')
'A generic type parameter.'

TAverage_co = TypeVar('TAverage_co', covariant=True)
'A generic covariant type parameter.'

TCollection = TypeVar('TCollection')
'A generic type parameter.'

TDefault = TypeVar('TDefault')
'A generic type parameter.'

TInner = TypeVar('TInner')
'A generic type parameter.'

TKey = TypeVar('TKey')
'A generic type parameter.'

TKey2 = TypeVar('TKey2')
'A generic type parameter.'

TKey_co = TypeVar('TKey_co', covariant=True)
'A generic covariant type parameter.'

TOther = TypeVar('TOther')
'A generic type parameter.'

TOther2 = TypeVar('TOther2')
'A generic type parameter.'

TOther3 = TypeVar('TOther3')
'A generic type parameter.'

TOther4 = TypeVar('TOther4')
'A generic type parameter.'

TResult = TypeVar('TResult')
'A generic type parameter.'

TSelf = TypeVar('TSelf')
'A generic type parameter.'

TSource = TypeVar('TSource')
'A generic type parameter.'

TSource_co = TypeVar('TSource_co', covariant=True)
'A generic covariant type parameter.'

TValue = TypeVar('TValue')
'A generic type parameter.'

TValue_co = TypeVar('TValue_co', covariant=True)
'A generic covariant type parameter.'


@runtime_checkable
class SupportsAverage(Protocol[TAverage_co]):
    '''
    Instances of this protocol supports the averaging operation. that is, if `x` is such an instance,
    and `N` is an integer, then `(x + x + ...) / N` is allowed, and has the type `TAverage_co`.
    '''
    @abstractmethod
    def __add__(self: TSelf, __o: TSelf) -> TSelf: ...
    @abstractmethod
    def __truediv__(self, __o: int) -> TAverage_co: ...


@runtime_checkable
class SupportsLessThan(Protocol, metaclass=ABCMeta):
    '''
    Instances of this protocol supports the `<` operation.

    Even though they may be unimplemented, the existence of `<` implies the existence of `>`,
    and probably `==`, `!=`, `<=` and `>=`.
    '''
    @abstractmethod
    def __lt__(self, __o: Any) -> bool: ...

@runtime_checkable
class SupportsAdd(Protocol, metaclass=ABCMeta):
    '''
    Instances of this protocol supports the homogeneous `+` operation.
    '''
    @abstractmethod
    def __add__(self: TSelf, __o: TSelf) -> TSelf: ...


# a type like 'Hashable' is rn useless because a subclass of a hashable class may not
# be hashable (object -> list)

TSupportsLessThan = TypeVar('TSupportsLessThan', bound=SupportsLessThan)
'A generic type parameter that represents a type that `SupportsLessThan`.'

TSupportsAdd = TypeVar('TSupportsAdd', bound=SupportsAdd)
'A generic type parameter that represents a type that `SupportsAdd`.'



from __future__ import annotations
from typing import Callable, Generic, Optional, overload

from .enumerable import Enumerable
from .more_typing import (
    TSource_co,
    TKey,
    TKey2,
    TSupportsLessThan,
)


class OrderedEnumerable(Enumerable[TSource_co], Generic[TSource_co, TKey]):
    '''
    ```py
    from types_linq.ordered_enumerable import OrderedEnumerable
    ```

    Represents a sorted Enumerable sequence that is sorted by some key.

    Users should not construct instances of this class directly. Use `Enumerable.order_by()` instead.
    '''

    def __init__(self, *args): ...

    def create_ordered_enumerable(self,
        key_selector: Callable[[TSource_co], TKey2],
        comparer: Optional[Callable[[TKey2, TKey2], int]],
        descending: bool,
    ) -> OrderedEnumerable[TSource_co, TKey2]:
        '''
        Performs a subsequent ordering on the elements of the sequence according to a key.

        Comparer takes two values and return positive ints when lhs > rhs, negative ints
        if lhs < rhs, and 0 if they are equal.

        Revisions
            ~ v0.1.2: Fixed incorrect parameter type of comparer.
        '''

    @overload
    def then_by(self,
        key_selector: Callable[[TSource_co], TSupportsLessThan],
    ) -> OrderedEnumerable[TSource_co, TSupportsLessThan]:
        '''
        Performs a subsequent ordering of the elements in ascending order according to key.

        Example
        ```py
        >>> class Pet(NamedTuple):
        ...     name: str
        ...     age: int

        >>> pets = [Pet('Barley', 8), Pet('Boots', 4), Pet('Roman', 5), Pet('Daisy', 4)]
        >>> Enumerable(pets).order_by(lambda p: p.age) \\
        ...     .then_by(lambda p: p.name)             \\
        ...     .select(lambda p: p.name)              \\
        ...     .to_list()
        ['Boots', 'Daisy', 'Roman', 'Barley']
        ```
        '''

    @overload
    def then_by(self,
        key_selector: Callable[[TSource_co], TKey2],
        __comparer: Callable[[TKey2, TKey2], int],
    ) -> OrderedEnumerable[TSource_co, TKey2]:
        '''
        Performs a subsequent ordering of the elements in ascending order by using a specified comparer.

        Such comparer takes two values and return positive ints when lhs > rhs, negative ints
        if lhs < rhs, and 0 if they are equal.
        '''

    @overload
    def then_by_descending(self,
        key_selector: Callable[[TSource_co], TSupportsLessThan],
    ) -> OrderedEnumerable[TSource_co, TSupportsLessThan]:
        '''
        Performs a subsequent ordering of the elements in descending order according to key.
        '''

    @overload
    def then_by_descending(self,
        key_selector: Callable[[TSource_co], TKey2],
        __comparer: Callable[[TKey2, TKey2], int],
    ) -> OrderedEnumerable[TSource_co, TKey2]:
        '''
        Performs a subsequent ordering of the elements in descending order by using a specified comparer.

        Such comparer takes two values and return positive ints when lhs > rhs, negative ints
        if lhs < rhs, and 0 if they are equal.
        '''



from typing import Callable, Generic, Iterable, Optional, overload

from .more_enumerable import MoreEnumerable
from ..enumerable import Enumerable
from ..more_typing import (
    TKey,
    TSource_co,
)


class ExtremaEnumerable(MoreEnumerable[TSource_co], Generic[TSource_co, TKey]):
    '''
    ```py
    from types_linq.more.extrema_enumerable import ExtremaEnumerable
    ```

    Specialization for manipulating extrema.

    Users should not construct instances of this class directly. Use `MoreEnumerable.maxima_by()`
    instead.

    Revisions
        ~ v0.2.0: New.
    '''

    def __init__(self,
        source: Iterable[TSource_co],
        selector: Callable[[TSource_co], TKey],
        comparer: Optional[Callable[[TKey, TKey], int]],
        for_min: bool,
    ) -> None: ...  # internal

    @overload
    def take(self, count: int) -> MoreEnumerable[TSource_co]:
        '''
        Returns a specified number of contiguous elements from the start of the sequence.
        '''

    @overload
    def take(self, __index: slice) -> Enumerable[TSource_co]:
        '''
        Identical to parent.

        Revisions
            ~ v1.1.0: Fixed incorrect override of `Enumerable.take()` when it takes a slice.
        '''

    def take_last(self, count: int) -> MoreEnumerable[TSource_co]:
        '''
        Returns a new sequence that contains the last `count` elements.
        '''



from __future__ import annotations
from typing import Any, Tuple, Iterator, Hashable, List, Dict, Iterable, Set, Optional, MutableMapping, MutableSet

from .more_typing import (
    TKey,
    TSupportsLessThan,
    TValue
)


def identity(x: Any) -> Any:
    return x


def return_second(x: Any, y: TValue) -> TValue:
    return y


def default_equal(x: TValue, y: TValue) -> bool:
    return x == y


def default_lt(x: TSupportsLessThan, y: TSupportsLessThan) -> bool:
    return x < y


def default_gt(x: TSupportsLessThan, y: TSupportsLessThan) -> bool:
    return y < x


class _ListMap(MutableMapping[TKey, TValue]):

    def __init__(self) -> None:
        self._lst: List[Tuple[TKey, TValue]] = list()
    
    def __setitem__(self, key: TKey, value: TValue) -> None:
        for i in range(0, len(self._lst)):
            if self._lst[i][0] == key:
                self._lst[i] = (self._lst[i][0], value)
                return
        self._lst.append((key, value))
    
    def __getitem__(self, key: TKey) -> TValue:
        for k, v in self._lst:
            if k == key:
                return v
        raise KeyError(key)

    def __delitem__(self, key: TKey) -> None:
        for i in range(0, len(self._lst)):
            if self._lst[i][0] == key:
                del self._lst[i]
                return
        raise KeyError(key)

    def __iter__(self) -> Iterator[TKey]:
        return map(lambda e: e[0], self._lst)

    def __len__(self) -> int:
        return len(self._lst)


# Wrap Map to Set
class _SetWrapper(MutableSet[TValue]):

    def __init__(self, map: MutableMapping[TValue, None]) -> None:
        self._map = map

    def __contains__(self, x: object) -> bool:
        return x in self._map

    def __len__(self) -> int:
        return self._map.__len__()

    def add(self, value: TValue) -> None:
        self._map[value] = None

    def discard(self, value: TValue) -> None:
        try:
            self._map.pop(value)
        except KeyError:
            # set.discard() don't throw KeyError 
            # if element doesn't exist 
            ...
    
    def __iter__(self) -> Iterator[TValue]:
        return self._map.__iter__()


class ComposeSet(MutableSet[TValue]):
    '''
    A set which support hashable and unhashable type
    '''
    _set: Set[TValue]
    _cmp_set: MutableSet[TValue]

    def __init__(self, iter: Optional[Iterable[TValue]] = None):
        self._set = set()
        self._cmp_set = _SetWrapper(_ListMap())
        if iter:
            for item in iter:
                self.add(item)

    def _get_set(self, x: object) -> MutableSet[TValue]:
        if isinstance(x, Hashable):
            return self._set
        else:
            return self._cmp_set

    def __contains__(self, x: object) -> bool:
        return x in self._get_set(x)

    def __len__(self) -> int:
        return len(self._set) + len(self._cmp_set)

    def __iter__(self) -> Iterator[TValue]:
        yield from self._set
        yield from self._cmp_set

    def add(self, value: TValue) -> None:
        return self._get_set(value).add(value)

    def discard(self, value: TValue) -> None:
        return self._get_set(value).discard(value)


class ComposeMap(MutableMapping[TKey, TValue]):
    '''
    A map which support hashable and unhashable key type
    '''
    _map: Dict[TKey, TValue]
    _cmp_map: MutableMapping[TKey, TValue]

    def __init__(self, iter: Optional[Iterable[Tuple[TKey, TValue]]] = None) -> None:
        self._map = dict()
        self._cmp_map = _ListMap()
        if iter:
            for k, v in iter:
                self[k] = v

    def _get_map(self, x: object) -> MutableMapping[TKey, TValue]:
        # TODO: this test is not perfectly sufficient.
        # counterexample: x is namedtuple('t', ['x'])([])
        if isinstance(x, Hashable):
            return self._map
        else:
            return self._cmp_map

    def __setitem__(self, key: TKey, value: TValue):
        self._get_map(key).__setitem__(key, value)
    
    def __getitem__(self, key: TKey) -> TValue:
        return self._get_map(key).__getitem__(key)

    def __delitem__(self, key: TKey) -> None:
        return self._get_map(key).__delitem__(key)

    def __iter__(self) -> Iterator[TKey]:
        # notice on iteration order:
        # if this data structure contains only hashable objects (stored in the dict)
        # then the iter order is consistent with dict => using insertion order. this is
        # useful for meeting the .net ToLookup() spec. however, if this is not the case
        # then the iter order is broken.
        yield from self._map
        yield from self._cmp_map

    def __len__(self) -> int:
        return len(self._map) + len(self._cmp_map)



from typing import Any, Callable, Iterable, Optional, Tuple, overload

from ..enumerable import Enumerable
from .extrema_enumerable import ExtremaEnumerable
from .more_enums import RankMethods
from ..more_typing import (
    TAccumulate,
    TKey,
    TResult,
    TSource,
    TSource_co,
    TSupportsLessThan,
)


class MoreEnumerable(Enumerable[TSource_co]):
    '''
    ```py
    from types_linq.more import MoreEnumerable
    ```

    MoreEnumerable provides more query methods. Instances of this class can be created by directly
    constructing, using `as_more()`, or invoking MoreEnumerable methods that return MoreEnumerable
    instead of Enumerable.

    These APIs may have breaking changes more frequently than those in Enumerable class because updates
    in .NET are happening and sometimes ones of these APIs could be moved to Enumerable with modification,
    or changed to accommodate changes to Enumerable.

    Revisions
        ~ v0.2.0: New.
    '''

    @overload
    def aggregate_right(self,
        __seed: TAccumulate,
        __func: Callable[[TSource_co, TAccumulate], TAccumulate],
        __result_selector: Callable[[TAccumulate], TResult],
    ) -> TResult:
        '''
        Applies a right-associative accumulator function over the sequence. The seed is used as
        the initial accumulator value, and the result_selector is used to select the result value.

        Revisions
            ~ v1.2.0: Fixed annotation for __func.
        '''

    @overload
    def aggregate_right(self,
        __seed: TAccumulate,
        __func: Callable[[TSource_co, TAccumulate], TAccumulate],
    ) -> TAccumulate:
        '''
        Applies a right-associative accumulator function over the sequence. The seed is used as the
        initial accumulator value.

        Example
        ```py
        >>> values = [9, 4, 2]
        >>> MoreEnumerable(values).aggregate_right('null', lambda e, rr: f'(cons {e} {rr})')
        '(cons 9 (cons 4 (cons 2 null)))'
        ```

        Revisions
            ~ v1.2.0: Fixed annotation for __func.
        '''

    @overload
    def aggregate_right(self,
        __func: Callable[[TSource_co, TSource_co], TSource_co],
    ) -> TSource_co:
        '''
        Applies a right-associative accumulator function over the sequence. Raises `InvalidOperationError`
        if there is no value in the sequence.

        Example
        ```py
        >>> values = ['9', '4', '2', '5']
        >>> MoreEnumerable(values).aggregate_right(lambda e, rr: f'({e}+{rr})')
        '(9+(4+(2+5)))'
        ```

        Revisions
            ~ v1.2.0: Fixed annotation for __func.
        '''

    def as_more(self) -> MoreEnumerable[TSource_co]:
        '''
        Returns the original MoreEnumerable reference.
        '''

    def consume(self) -> None:
        '''
        Consumes the sequence completely. This method iterates the sequence immediately and does not save
        any intermediate data.

        Revisions
            ~ v1.1.0: New.
        '''

    def cycle(self, count: Optional[int] = None) -> MoreEnumerable[TSource_co]:
        '''
        Repeats the sequence `count` times.

        If `count` is `None`, the sequence is infinite. Raises `InvalidOperationError` if `count`
        is negative.

        Example
        ```py
        >>> MoreEnumerable([1, 2, 3]).cycle(3).to_list()
        [1, 2, 3, 1, 2, 3, 1, 2, 3]
        ```

        Revisions
            ~ v1.1.0: New.
        '''

    def enumerate(self, start: int = 0) -> MoreEnumerable[Tuple[int, TSource_co]]:
        '''
        Returns a sequence of tuples containing the index and the value from the source sequence. `start`
        is used to specify the starting index.

        Example
        ```py
        >>> ints = [2, 4, 6]
        >>> MoreEnumerable(ints).enumerate().to_list()
        [(0, 2), (1, 4), (2, 6)]
        ```

        Revisions
            ~ v1.0.0: New.
        '''

    def except_by2(self,
        second: Iterable[TSource_co],
        key_selector: Callable[[TSource_co], object],
    ) -> MoreEnumerable[TSource_co]:
        '''
        Produces the set difference of two sequences: self - second, according to a key selector that
        determines "distinctness". Note the second iterable is homogenous to self.

        Example
        ```py
        >>> first = [(16, 'x'), (9, 'y'), (12, 'd'), (16, 't')]
        >>> second = [(24, 'd'), (77, 'y')]
        >>> MoreEnumerable(first).except_by2(second, lambda x: x[1]).to_list()
        [(16, 'x'), (16, 't')]
        ```

        Revisions
            ~ v1.0.0: Renamed from `except_by()` to this name to accommodate an update to Enumerable class.
            ~ v0.2.1: Added preliminary support for unhashable keys.
        '''

    @overload
    def flatten(self) -> MoreEnumerable[Any]:
        '''
        Flattens the sequence containing arbitrarily-nested subsequences.

        Note: the nested objects must be Iterable to be flatten.
        Instances of `str` or `bytes` are not flattened.

        Example
        ```py
        >>> lst = ['apple', ['orange', ['juice', 'mango'], 'delta function']]
        >>> MoreEnumerable(lst).flatten().to_list()
        ['apple', 'orange', 'juice', 'mango', 'delta function']
        ```
        '''

    @overload
    def flatten(self, __predicate: Callable[[Iterable[Any]], bool]) -> MoreEnumerable[Any]:
        '''
        Flattens the sequence containing arbitrarily-nested subsequences. A predicate function determines
        whether a nested iterable should be flattened or not.

        Note: the nested objects must be Iterable to be flatten.
        '''

    def flatten2(self, selector: Callable[[Any], Optional[Iterable[object]]]) -> MoreEnumerable[Any]:
        '''
        Flattens the sequence containing arbitrarily-nested subsequences. A selector is used to select a
        subsequence based on the object's properties. If the selector returns None, then the object is
        considered a leaf.
        '''

    def for_each(self, action: Callable[[TSource_co], object]) -> None:
        '''
        Executes the given function on each element in the source sequence. The return values are discarded.

        Example
        ```py
        >>> def gen():
        ...     yield 116; yield 35; yield -9

        >>> Enumerable(gen()).where(lambda x: x > 0).as_more().for_each(print)
        116
        35
        ```
        '''

    def for_each2(self, action: Callable[[TSource_co, int], object]) -> None:
        '''
        Executes the given function on each element in the source sequence. Each element's index is used in
        the logic of the function. The return values are discarded.
        '''

    def interleave(self, *iters: Iterable[TSource_co]) -> MoreEnumerable[TSource_co]:
        '''
        Interleaves the elements of two or more sequences into a single sequence, skipping sequences if they
        are consumed.

        Example
        ```py
        >>> MoreEnumerable(['1', '2']).interleave(['4', '5', '6'], ['7', '8', '9']).to_list()
        ['1', '4', '7', '2', '5', '8', '6', '9']
        ```
        '''

    @overload
    def maxima_by(self,
        selector: Callable[[TSource_co], TSupportsLessThan],
    ) -> ExtremaEnumerable[TSource_co, TSupportsLessThan]:
        '''
        Returns the maximal elements of the sequence based on the given selector.

        Example
        ```py
        >>> strings = ['foo', 'bar', 'cheese', 'orange', 'baz', 'spam', 'egg', 'toasts', 'dish']
        >>> MoreEnumerable(strings).maxima_by(len).to_list()
        ['cheese', 'orange', 'toasts']
        >>> MoreEnumerable(strings).maxima_by(lambda x: x.count('e')).first()
        'cheese'
        ```
        '''

    @overload
    def maxima_by(self,
        selector: Callable[[TSource_co], TKey],
        __comparer: Callable[[TKey, TKey], int],
    ) -> ExtremaEnumerable[TSource_co, TKey]:
        '''
        Returns the maximal elements of the sequence based on the given selector and the comparer.

        Such comparer takes two values and return positive ints when lhs > rhs, negative ints
        if lhs < rhs, and 0 if they are equal.
        '''

    @overload
    def minima_by(self,
        selector: Callable[[TSource_co], TSupportsLessThan],
    ) -> ExtremaEnumerable[TSource_co, TSupportsLessThan]:
        '''
        Returns the minimal elements of the sequence based on the given selector.
        '''

    @overload
    def minima_by(self,
        selector: Callable[[TSource_co], TKey],
        __comparer: Callable[[TKey, TKey], int],
    ) -> ExtremaEnumerable[TSource_co, TKey]:
        '''
        Returns the minimal elements of the sequence based on the given selector and the comparer.

        Such comparer takes two values and return positive ints when lhs > rhs, negative ints
        if lhs < rhs, and 0 if they are equal.
        '''

    def pipe(self, action: Callable[[TSource_co], object]) -> MoreEnumerable[TSource_co]:
        '''
        Executes the given action on each element in the sequence and yields it. Return values of
        action are discarded.

        Example
        ```py
        >>> store = set()
        >>> MoreEnumerable([1, 2, 2, 1]).pipe(store.add).where(lambda x: x % 2 == 0).to_list()
        [2, 2]
        >>> store
        {1, 2}
        ```

        Revisions
            ~ v0.2.1: New.
        '''

    # note: diffrent from morelinq: identity is first parameter
    def pre_scan(self,
        identity: TAccumulate,
        transformation: Callable[[TAccumulate, TSource_co], TAccumulate],
    ) -> MoreEnumerable[TAccumulate]:
        '''
        Performs a pre-scan (exclusive prefix sum) over the sequence. Such scan returns an
        equal-length sequence where the first element is the identity, and i-th element (i>1) is
        the sum of the first i-1 (and identity) elements in the original sequence.

        Example
        ```py
        >>> values = [9, 4, 2, 5, 7]
        >>> MoreEnumerable(values).pre_scan(0, lambda acc, e: acc + e).to_list()
        [0, 9, 13, 15, 20]
        >>> MoreEnumerable([]).pre_scan(0, lambda acc, e: acc + e).to_list()
        []
        ```

        Revisions
            ~ v1.2.0: New.
        '''

    @overload
    def rank(self: MoreEnumerable[TSupportsLessThan],
        *,
        method: RankMethods = RankMethods.dense,
    ) -> MoreEnumerable[int]:
        '''
        Ranks each item in the sequence in descending order using the method provided.

        Example
        ```py
        >>> scores = [1, 4, 77, 23, 23, 4, 9, 0, -7, 101, 23]
        >>> MoreEnumerable(scores).rank().to_list()
        [6, 5, 2, 3, 3, 5, 4, 7, 8, 1, 3]  # 101 is largest, so has rank of 1

        >>> MoreEnumerable(scores).rank(method=RankMethods.competitive).to_list()
        [9, 7, 2, 3, 3, 7, 6, 10, 11, 1, 3]  # there are no 4th or 5th since there
                                             # are three 3rd's

        >>> MoreEnumerable(scores).rank(method=RankMethods.ordinal).to_list()
        [9, 7, 2, 3, 4, 8, 6, 10, 11, 1, 5]  # as in sorting
        ```

        Revisions
            ~ v1.2.1: Added method parameter to support more ranking methods.
            ~ v1.0.0: New.
        '''

    @overload
    def rank(self,
        __comparer: Callable[[TSource_co, TSource_co], int],
        *,
        method: RankMethods = RankMethods.dense,
    ) -> MoreEnumerable[int]:
        '''
        Ranks each item in the sequence in descending order using the given comparer and the
        method.

        Such comparer takes two values and return positive ints when lhs > rhs, negative ints
        if lhs < rhs, and 0 if they are equal.

        Revisions
            ~ v1.2.1: Added method parameter to support more ranking methods.
            ~ v1.0.0: New.
        '''

    @overload
    def rank_by(self,
        key_selector: Callable[[TSource_co], TSupportsLessThan],
        *,
        method: RankMethods = RankMethods.dense,
    ) -> MoreEnumerable[int]:
        '''
        Ranks each item in the sequence in descending order using the given selector and the
        method.

        Example
        ```py
        >>> scores = [
        ...     {'name': 'Frank', 'score': 75},
        ...     {'name': 'Alica', 'score': 90},
        ...     {'name': 'Erika', 'score': 99},
        ...     {'name': 'Rogers', 'score': 90},
        ... ]

        >>> MoreEnumerable(scores).rank_by(lambda x: x['score']) \\
        ...     .zip(scores) \\
        ...     .group_by(lambda t: t[0], lambda t: t[1]['name']) \\
        ...     .to_dict(lambda g: g.key, lambda g: g.to_list())
        {3: ['Frank'], 2: ['Alica', 'Rogers'], 1: ['Erika']}
        ```

        Revisions
            ~ v1.2.1: Added method parameter to support more ranking methods.
            ~ v1.0.0: New.
        '''

    @overload
    def rank_by(self,
        key_selector: Callable[[TSource_co], TKey],
        __comparer: Callable[[TKey, TKey], int],
        *,
        method: RankMethods = RankMethods.dense,
    ) -> MoreEnumerable[int]:
        '''
        Ranks each item in the sequence in descending order using the given selector, comparer
        and the method.

        Such comparer takes two values and return positive ints when lhs > rhs, negative ints
        if lhs < rhs, and 0 if they are equal.

        Revisions
            ~ v1.2.1: Added method parameter to support more ranking methods.
            ~ v1.0.0: New.
        '''

    @overload
    def run_length_encode(self) -> MoreEnumerable[Tuple[TSource_co, int]]:
        '''
        Run-length encodes the sequence into a sequence of tuples where each tuple contains an
        (the first) element and its number of contingent occurrences, where equality is based on
        `==`.

        Example
        ```py
        >>> MoreEnumerable('abbcaeeeaa').run_length_encode().to_list()
        [('a', 1), ('b', 2), ('c', 1), ('a', 1), ('e', 3), ('a', 2)]
        ```

        Revisions
            ~ v1.1.0: New.
        '''

    @overload
    def run_length_encode(self,
        __comparer: Callable[[TSource_co, TSource_co], bool],
    ) -> MoreEnumerable[Tuple[TSource_co, int]]:
        '''
        Run-length encodes the sequence into a sequence of tuples where each tuple contains an
        (the first) element and its number of contingent occurrences, where equality is determined by
        the comparer.

        Example
        ```py
        >>> MoreEnumerable('abBBbcaEeeff') \\
        >>>     .run_length_encode(lambda x, y: x.lower() == y.lower()).to_list()
        [('a', 1), ('b', 4), ('c', 1), ('a', 1), ('E', 3), ('f', 2)]
        ```

        Revisions
            ~ v1.1.0: New.
        '''

    @overload
    def scan(self,
        __transformation: Callable[[TSource_co, TSource_co], TSource_co],
    ) -> MoreEnumerable[TSource_co]:
        '''
        Performs a inclusive prefix sum over the sequence. Such scan returns an equal-length sequence
        where the i-th element is the sum of the first i elements in the original sequence.

        Example
        ```py
        >>> values = [9, 4, 2, 5, 7]
        >>> MoreEnumerable(values).scan(lambda acc, e: acc + e).to_list()
        [9, 13, 15, 20, 27]
        >>> MoreEnumerable([]).scan(lambda acc, e: acc + e).to_list()
        []
        ```

        Example
        ```py
        >>> # running max
        >>> fruits = ['apple', 'mango', 'orange', 'passionfruit', 'grape']
        >>> MoreEnumerable(fruits).scan(lambda acc, e: e if len(e) > len(acc) else acc).to_list()
        ['apple', 'apple', 'orange', 'passionfruit', 'passionfruit']
        ```

        Revisions
            ~ v1.2.0: New.
        '''

    @overload
    def scan(self,
        __seed: TAccumulate,
        __transformation: Callable[[TAccumulate, TSource_co], TAccumulate],
    ) -> MoreEnumerable[TAccumulate]:
        '''
        Like Enumerable.aggregate(seed, transformation) except that the intermediate results are
        included in the result sequence.

        Example
        ```py
        >>> Enumerable.range(1, 5).as_more().scan(-1, lambda acc, e: acc * e).to_list()
        [-1, -1, -2, -6, -24, -120]
        ```

        Revisions
            ~ v1.2.0: New.
        '''

    @overload
    def scan_right(self,
        __func: Callable[[TSource_co, TSource_co], TSource_co],
    ) -> MoreEnumerable[TSource_co]:
        '''
        Performs a right-associative inclusive prefix sum over the sequence. This is the
        right-associative version of MoreEnumerable.scan(func).

        Example
        ```py
        >>> values = ['9', '4', '2', '5']
        >>> MoreEnumerable(values).scan_right(lambda e, rr: f'({e}+{rr})').to_list()
        ['(9+(4+(2+5)))', '(4+(2+5))', '(2+5)', '5']
        >>> MoreEnumerable([]).scan_right(lambda e, rr: e + rr).to_list()
        []
        ```

        Revisions
            ~ v1.2.0: New.
        '''

    @overload
    def scan_right(self,
        __seed: TAccumulate,
        __func: Callable[[TSource_co, TAccumulate], TAccumulate],
    ) -> MoreEnumerable[TAccumulate]:
        '''
        The right-associative version of MoreEnumerable.scan(seed, func).

        Example
        ```py
        >>> values = [9, 4, 2]
        >>> MoreEnumerable(values).scan_right('null', lambda e, rr: f'(cons {e} {rr})').to_list()
        ['(cons 9 (cons 4 (cons 2 null)))', '(cons 4 (cons 2 null))', '(cons 2 null)', 'null']
        ```

        Revisions
            ~ v1.2.0: New.
        '''

    def segment(self,
        new_segment_predicate: Callable[[TSource_co], bool],
    ) -> MoreEnumerable[MoreEnumerable[TSource_co]]:
        '''
        Splits the sequence into segments by using a detector function that returns True to signal a
        new segment.

        Example
        ```py
        >>> values = [0, 1, 2, 4, -4, -2, 6, 2, -2]
        >>> MoreEnumerable(values).segment(lambda x: x < 0).select(lambda x: x.to_list()).to_list()
        [[0, 1, 2, 4], [-4], [-2, 6, 2], [-2]]
        ```

        Revisions
            ~ v1.2.0: New.
        '''

    def segment2(self,
        new_segment_predicate: Callable[[TSource_co, int], bool],
    ) -> MoreEnumerable[MoreEnumerable[TSource_co]]:
        '''
        Splits the sequence into segments by using a detector function that returns True to signal a
        new segment. The element's index is used in the detector function.

        Example
        ```py
        >>> values = [0, 1, 2, 4, -4, -2, 6, 2, -2]
        >>> MoreEnumerable(values).segment2(lambda x, i: x < 0 or i % 3 == 0) \\
        ...     .select(lambda x: x.to_list()) \\
        ...     .to_list()
        [[0, 1, 2], [4], [-4], [-2], [6, 2], [-2]]
        ```

        Revisions
            ~ v1.2.0: New.
        '''

    def segment3(self,
        new_segment_predicate: Callable[[TSource_co, TSource_co, int], bool],
    ) -> MoreEnumerable[MoreEnumerable[TSource_co]]:
        '''
        Splits the sequence into segments by using a detector function that returns True to signal a
        new segment. The last element and the current element's index are used in the detector
        function.

        Example
        ```py
        >>> values = [0, 1, 2, 4, -4, -2, 6, 2, -2]
        >>> MoreEnumerable(values).segment3(lambda curr, prev, i: curr * prev < 0) \\
        ...     .select(lambda x: x.to_list()) \\
        ...     .to_list()
        [[0, 1, 2, 4], [-4, -2], [6, 2], [-2]]
        ```

        Revisions
            ~ v1.2.0: New.
        '''

    @staticmethod
    def traverse_breath_first(
        root: TSource,
        children_selector: Callable[[TSource], Iterable[TSource]],
    ) -> MoreEnumerable[TSource]:
        '''
        Traverses the tree (graph) from the root node in a breath-first fashion. A selector is used to
        select children of each node.

        Graphs are not checked for cycles or duplicates visits. If the resulting sequence needs to be
        finite then it is the responsibility of children_selector to ensure that duplicate nodes are not
        visited.

        Example
        ```py
        >>> tree = { 3: [1, 4], 1: [0, 2], 4: [5] }
        >>> MoreEnumerable.traverse_breath_first(3, lambda x: tree.get(x, [])) \\
        >>>     .to_list()
        [3, 1, 4, 0, 2, 5]
        ```
        '''

    @staticmethod
    def traverse_depth_first(
        root: TSource,
        children_selector: Callable[[TSource], Iterable[TSource]],
    ) -> MoreEnumerable[TSource]:
        '''
        Traverses the tree (graph) from the root node in a depth-first fashion. A selector is used to
        select children of each node.

        Graphs are not checked for cycles or duplicates visits. If the resulting sequence needs to be
        finite then it is the responsibility of children_selector to ensure that duplicate nodes are not
        visited.

        Example
        ```py
        >>> tree = { 3: [1, 4], 1: [0, 2], 4: [5] }
        >>> MoreEnumerable.traverse_depth_first(3, lambda x: tree.get(x, [])) \\
        >>>     .to_list()
        [3, 1, 0, 2, 4, 5]
        ```
        '''

    # NOTE: leave room for comparer overloads
    # TODO: when hasher support is done
    def traverse_topological(self,
        children_selector: Callable[[TSource_co], Iterable[TSource_co]],
    ) -> MoreEnumerable[TSource_co]:
        '''
        Traverses the graph in topological order, A selector is used to select children of each
        node. The ordering created from this method is a variant of depth-first traversal and ensures
        duplicate nodes are output once.

        To invoke this method, the self sequence contains nodes with zero in-degrees to start the
        iteration. Passing a list of all nodes is allowed although not required.

        Raises `DirectedGraphNotAcyclicError` if the directed graph contains a cycle and the
        topological ordering cannot be produced.

        Example
        ```py
        >>> adj = { 5: [2, 0], 4: [0, 1], 2: [3], 3: [1] }
        >>> MoreEnumerable([5, 4]).traverse_topological(lambda x: adj.get(x, [])) \\
        >>>     .to_list()
        [5, 2, 3, 4, 0, 1]
        ```

        Revisions
            ~ v1.2.1: New.
        '''

    def traverse_topological2(self,
        children_selector: Callable[[TSource_co], Iterable[TSource_co]],
        key_selector: Callable[[TSource_co], object],
    ) -> MoreEnumerable[TSource_co]:
        '''
        Traverses the graph in topological order, A selector is used to select children of each
        node. The ordering created from this method is a variant of depth-first traversal and
        ensures duplicate nodes are output once. A key selector is used to determine equality
        between nodes.

        To invoke this method, the self sequence contains nodes with zero in-degrees to start the
        iteration. Passing a list of all nodes is allowed although not required.

        Raises `DirectedGraphNotAcyclicError` if the directed graph contains a cycle and the
        topological ordering cannot be produced.

        Revisions
            ~ v1.2.1: New.
        '''



from __future__ import annotations
from typing import Any, Callable, Container, Deque, Dict, Iterable, Iterator, List, MutableSequence, NoReturn, Optional, Reversible, Sequence, Set, Sized, TYPE_CHECKING, Tuple, Type, Generic, Union

if TYPE_CHECKING:
    from .lookup import Lookup
    from .grouping import Grouping
    from .ordered_enumerable import OrderedEnumerable
    from .cached_enumerable import CachedEnumerable
    from .more import MoreEnumerable

from .types_linq_error import InvalidOperationError, IndexOutOfRangeError
from .util import (
    ComposeSet,
    default_equal,
    default_gt,
    default_lt,
    identity,
    return_second,
)
from .more_typing import (
    TCollection,
    TDefault,
    TInner,
    TKey,
    TResult,
    TSource_co,
    TValue,
)


# do not use this value!!!
_signal: Any = object()


class Enumerable(Sequence[TSource_co], Generic[TSource_co]):

    _iter_factory: Callable[[], Iterable[TSource_co]]

    def __init__(self,
        it: Union[Iterable[TSource_co], Callable[[], Iterable[TSource_co]]]
    ):
        if isinstance(it, Iterable):
            self._iter_factory = lambda it_=it: it_
        else:
            self._iter_factory = lambda it_=it: it_()

    def _get_iterable(self) -> Iterable[TSource_co]:
        return self._iter_factory()

    # 'fallback = F' -> calls dunder methods if available
    #                  -> otherwise calls own implementation
    # 'fallback = T'-> calls own implementation
    def _contains_impl(self, value: object, fallback: bool) -> bool:
        iterable = self._get_iterable()
        if not fallback and isinstance(iterable, Container):
            return value in iterable  # type: ignore
        for elem in iterable:
            if elem == value:
                return True
        return False

    def __contains__(self, value: object) -> bool:
        return self._contains_impl(value, fallback=False)

    @staticmethod
    def _raise_not_enough_elements() -> NoReturn:
        raise IndexOutOfRangeError('Not enough elements in the sequence')

    def _every(self, step: int) -> Enumerable[TSource_co]:
        return self.where2(lambda _, i: i % step == 0)

    def _getitem_impl(self,
        index: Union[int, slice],
        fallback: bool,
    ) -> Union[TSource_co, Enumerable[TSource_co]]:
        iterable = self._get_iterable()
        if isinstance(index, int):
            # Sequence is an abstract base class without @runtime_checkable
            if not fallback and isinstance(iterable, Sequence):
                # an appropriate implementation should raise IndexError, or IndexOutOfRangeError
                try:
                    return iterable[index]
                except IndexError as e:
                    raise IndexOutOfRangeError from e
            if index >= 0:
                iterator = iter(iterable)
                try:
                    for _ in range(index):
                        next(iterator)
                    return next(iterator)
                except StopIteration:
                    self._raise_not_enough_elements()
            else:
                en = iterable if isinstance(iterable, Enumerable) else Enumerable(iterable)
                last = en.take_last(-index).to_list()
                if len(last) < -index:
                    self._raise_not_enough_elements()
                return last[0]

        else:  # isinstance(index, slice)
            if not fallback and isinstance(iterable, Sequence):
                try:
                    res = iterable[index]
                except IndexError as e:
                    raise IndexOutOfRangeError from e
                return res if isinstance(res, Enumerable) else Enumerable(res)
            # we do not enumerate all values if the begin and the end only involve
            # nonnegative indices since in which case the sliced part can be obtained
            # without reversing.
            # otherwise have to enumerate all with using list's slice operator.
            # (don't enumerate right away in this function, of course)
            def inner(s: slice = index):
                en = iterable if isinstance(iterable, Enumerable) else Enumerable(iterable)
                start_is_none = s.start is None
                stop_is_none = s.stop is None
                step = s.step if s.step is not None else 1
                if (start_is_none and stop_is_none) or (not start_is_none and s.start < 0) \
                    or (not stop_is_none and s.stop < 0) or (stop_is_none):
                    yield from en.to_list()[s]
                    return
                elif start_is_none:
                    if step > 0:
                        yield from en.take(s.stop)._every(step)
                    else:
                        yield from en.skip(s.stop + 1).reverse()._every(-step)
                    return
                elif s.start <= s.stop and step > 0:
                    yield from en.skip(s.start).take(s.stop - s.start)._every(step)
                    return
                elif step <= 0:
                    yield from en.skip(s.stop + 1).take(s.start - s.stop) \
                        .reverse()._every(-step)
            return Enumerable(inner)

    def __getitem__(self,  # type: ignore[override]
        index: Union[int, slice, Tuple[int, TDefault]],
    ) -> Union[TSource_co, Enumerable[TSource_co], TDefault]:
        if isinstance(index, tuple):
            idx, default = index
            try:
                return self._getitem_impl(idx, fallback=False)
            except IndexOutOfRangeError:
                return default
        else:  # isinstance(index, (int, slice))
            return self._getitem_impl(index, fallback=False)

    def __iter__(self) -> Iterator[TSource_co]:
        return iter(self._get_iterable())

    def _len_impl(self, fallback: bool) -> int:
        iterable = self._get_iterable()
        if not fallback and isinstance(iterable, Sized):
            return len(iterable)
        count = 0
        for _ in iterable: count += 1
        return count

    def __len__(self) -> int:
        return self._len_impl(fallback=False)

    def _reversed_impl(self, fallback: bool) -> Iterator[TSource_co]:
        iterable = self._get_iterable()
        # Sequence is an abstract base class without @runtime_checkable
        if not fallback and isinstance(iterable, (Sequence, Reversible)):
            return reversed(iterable)
        return reversed([elem for elem in iterable])

    def __reversed__(self) -> Iterator[TSource_co]:
        return self._reversed_impl(fallback=False)

    @staticmethod
    def _raise_empty_sequence() -> NoReturn:
        raise InvalidOperationError('Sequence is empty')

    def aggregate(self, *args) -> Any:
        if len(args) == 3:
            seed, func, result_selector = args
            for elem in self:
                seed = func(seed, elem)
            return result_selector(seed)

        elif len(args) == 2:
            seed, func = args
            for elem in self:
                seed = func(seed, elem)
            return seed

        else:  # len(args) == 1
            func = args[0]
            iterator = iter(self)
            try:
                seed = next(iterator)
            except StopIteration:
                self._raise_empty_sequence()
            for elem in iterator:
                seed = func(seed, elem)
            return seed

    def all(self, predicate: Callable[[TSource_co], bool]) -> bool:
        for elem in self:
            if not predicate(elem):
                return False
        return True

    def any(self, *args: Callable[[TSource_co], bool]) -> bool:
        if len(args) == 0:
            for _ in self:
                return True
            return False

        else:  # len(args) == 1:
            predicate = args[0]
            for elem in self:
                if predicate(elem):
                    return True
            return False

    def append(self, element: TSource_co) -> Enumerable[TSource_co]:  # type: ignore
        # this method does not mutate the current container
        def inner():
            yield from self
            yield element
        return Enumerable(inner)

    def as_cached(self, *, cache_capacity: Optional[int] = None) -> CachedEnumerable[TSource_co]:
        from .cached_enumerable import CachedEnumerable
        return CachedEnumerable(self, cache_capacity)

    def as_more(self) -> MoreEnumerable[TSource_co]:
        from .more import MoreEnumerable
        return MoreEnumerable(self)

    def _average_helper(self, selector, when_empty):
        count = 0
        iterator = iter(self)

        try:
            sum_ = selector(next(iterator))
            count += 1
        except StopIteration:
            return when_empty()

        for elem in iterator:
            sum_ += selector(elem)
            count += 1
        return sum_ / count

    def average(self, *args: Callable[[TSource_co], Any]) -> Any:
        if len(args) == 0:
            selector = identity
        else: # len(args) == 1
            selector = args[0]
        return self._average_helper(selector, self._raise_empty_sequence)

    def average2(self, *args):
        if len(args) == 1:
            selector, default = identity, args[0]
        else: # len(args) == 2
            selector, default = args
        return self._average_helper(selector, lambda: default)

    def cast(self, _: Type[TResult]) -> Enumerable[TResult]:
        return self  # type: ignore

    def chunk(self, size: int) -> Enumerable[MutableSequence[TSource_co]]:
        if size < 1:
            raise InvalidOperationError('size must be greater than 0')
        def inner():
            lst: List[Any] = [1] * size
            i = 0
            for elem in self:
                lst[i] = elem
                if i == size - 1:
                    yield lst
                    i = 0
                    lst = [1] * size
                else:
                    i += 1
            if i > 0:
                del lst[i:]
                yield lst
        return Enumerable(inner)

    def concat(self, second: Iterable[TSource_co]) -> Enumerable[TSource_co]:
        def inner():
            yield from self
            yield from second
        return Enumerable(inner)

    def contains(self, value: object, *args: Callable[..., bool]):
        if len(args) == 0:
            return self._contains_impl(value, fallback=True)

        else:  # len(args) == 1
            comparer = args[0]
            for elem in self:
                if comparer(elem, value):
                    return True
            return False

    def count(self, *args: Callable[[TSource_co], bool]) -> int:
        if len(args) == 0:
            return self._len_impl(fallback=True)

        else:  # len(args) == 1
            predicate = args[0]
            count = 0
            for elem in self:
                if predicate(elem):
                    count += 1
            return count

    def default_if_empty(self,
        default: TDefault,
    ) -> Union[Enumerable[TSource_co], Enumerable[TDefault]]:
        def inner():
            iterator = iter(self)
            try:
                yield next(iterator)
            except StopIteration:
                yield default
                return
            yield from iterator
        return Enumerable(inner)  # type: ignore

    def distinct(self) -> Enumerable[TSource_co]:
        return self.except1(())

    def distinct_by(self, key_selector: Callable[[TSource_co], object]) -> Enumerable[TSource_co]:
        return self.except_by((), key_selector)

    def element_at(self, index: int, *args: TDefault) -> Union[TSource_co, TDefault]:
        if len(args) == 0:
            return self._getitem_impl(index, fallback=True)  # type: ignore
        else:  # len(args) == 1
            default = args[0]
            try:
                return self._getitem_impl(index, fallback=True)  # type: ignore
            except IndexOutOfRangeError:
                return default

    @staticmethod
    def empty() -> Enumerable[TSource_co]:
        return Enumerable(())

    def except1(self, second: Iterable[TSource_co]) -> Enumerable[TSource_co]:
        return self.except_by(second, identity)

    def except_by(self,
        second: Iterable[TKey],
        key_selector: Callable[[TSource_co], TKey],
    ) -> Enumerable[TSource_co]:
        def inner():
            s = ComposeSet(second)
            for elem in self:
                key = key_selector(elem)
                if key in s:
                    continue
                s.add(key)
                yield elem
        return Enumerable(inner)

    @staticmethod
    def _raise_no_such_element() -> NoReturn:
        raise InvalidOperationError('No element satisfying condition')

    def first(self, *args: Callable[[TSource_co], bool]) -> TSource_co:
        if len(args) == 0:
            try:
                return self.element_at(0)  # type: ignore
            except IndexOutOfRangeError as e:
                raise InvalidOperationError(e)

        else:  # len(args) == 1
            predicate = args[0]
            for elem in self:
                if predicate(elem):
                    return elem
            self._raise_no_such_element()

    def first2(self, *args):
        if len(args) == 1:
            default = args[0]
            try:
                return self.element_at(0)  # type: ignore
            except IndexOutOfRangeError:
                return default

        else:  # len(args) == 2
            predicate, default = args
            for elem in self:
                if predicate(elem):
                    return elem
            return default

    def group_by(self,
        key_selector: Callable[[TSource_co], TKey],
        value_selector: Callable[[TSource_co], TValue],
        *args: Callable[[TKey, Enumerable[TValue]], TResult],
    ) -> Union[Enumerable[TResult], Enumerable[Grouping[TKey, TValue]]]:
        from .lookup import Lookup
        if len(args) == 1:
            result_selector = args[0]
            inner = lambda: Lookup(self, key_selector, value_selector) \
                .apply_result_selector(result_selector)  # type: ignore
            return Enumerable(inner)
        else:  # len(args) == 0:
            inner = lambda: Lookup(self, key_selector, value_selector)
            return Enumerable(inner)

    def group_by2(self,
        key_selector: Callable[[TSource_co], TKey],
        *args: Callable[[TKey, Enumerable[TSource_co]], TResult],
    ) -> Union[Enumerable[TResult], Enumerable[Grouping[TKey, TSource_co]]]:
        from .lookup import Lookup
        if len(args) == 1:
            result_selector = args[0]
            inner = lambda: Lookup(self, key_selector, identity) \
                .apply_result_selector(result_selector)  # type: ignore
            return Enumerable(inner)
        else:  # len(args) == 0:
            inner = lambda: Lookup(self, key_selector, identity)
            return Enumerable(inner)

    def group_join(self,
        inner: Iterable[TInner],
        outer_key_selector: Callable[[TSource_co], TKey],
        inner_key_selector: Callable[[TInner], TKey],
        result_selector: Callable[[TSource_co, Enumerable[TInner]], TResult],
    ) -> Enumerable[TResult]:
        from .lookup import Lookup
        def inner_gen():
            lookup = Lookup(inner, inner_key_selector, identity)
            for outer_item in self:
                group = lookup[outer_key_selector(outer_item)]
                yield result_selector(outer_item, group)  # type: ignore
        return Enumerable(inner_gen)

    def intersect(self, second: Iterable[TSource_co]) -> Enumerable[TSource_co]:
        return self.intersect_by(second, identity)

    def intersect_by(self,
        second: Iterable[TKey],
        key_selector: Callable[[TSource_co], TKey],
    ) -> Enumerable[TSource_co]:
        def inner():
            s = ComposeSet(second)
            for elem in self:
                key = key_selector(elem)
                if key not in s:
                    continue
                s.remove(key)
                yield elem
        return Enumerable(inner)

    def join(self,
        inner: Iterable[TInner],
        outer_key_selector: Callable[[TSource_co], TKey],
        inner_key_selector: Callable[[TInner], TKey],
        result_selector: Callable[[TSource_co, TInner], TResult],
    ) -> Enumerable[TResult]:
        from .lookup import Lookup
        def inner_gen():
            lookup = Lookup(inner, inner_key_selector, identity)
            for outer_item in self:
                for inner_item in lookup[outer_key_selector(outer_item)]:
                    yield result_selector(outer_item, inner_item)
        return Enumerable(inner_gen)

    def last(self, *args: Callable[[TSource_co], bool]) -> TSource_co:
        ret: Any = _signal
        if len(args) == 0:
            for elem in self:
                ret = elem
            if ret is _signal:
                self._raise_empty_sequence()

        else:  # len(args) == 1
            predicate = args[0]
            for elem in self:
                if predicate(elem):
                    ret = elem
            if ret is _signal:
                self._raise_no_such_element()
        return ret

    def last2(self, *args):
        if len(args) == 1:
            default = args[0]
            for elem in self:
                default = elem

        else:  # len(args) == 2
            predicate, default = args
            for elem in self:
                if predicate(elem):
                    default = elem
        return default

    def _minmax_helper(self, result_selector, op, when_empty) -> Any:
        iterator = iter(self)
        try:
            curr = result_selector(next(iterator))
        except StopIteration:
            return when_empty()
        for elem in iterator:
            mapped = result_selector(elem)
            curr = mapped if op(curr, mapped) else curr
        return curr

    def _minmax_by_helper(self, key_selector, op) -> Any:
        iterator = iter(self)
        try:
            curr = next(iterator)
        except StopIteration:
            self._raise_empty_sequence()
        curr_key = key_selector(curr)
        for elem in iterator:
            elem_key = key_selector(elem)
            if op(curr_key, elem_key):
                curr = elem
                curr_key = elem_key
        return curr

    def max(self, *args: Callable[[TSource_co], Any]) -> Any:
        if len(args) == 0:
            result_selector: Any = identity
        else:  # len(args) == 1
            result_selector = args[0]
        return self._minmax_helper(
            result_selector,
            default_lt,
            self._raise_empty_sequence,
        )

    def max2(self, *args) -> Any:
        if len(args) == 1:
            result_selector, default = identity, args[0]
        else:  # len(args) == 2
            result_selector, default = args
        return self._minmax_helper(
            result_selector,
            default_lt,
            lambda: default,
        )

    def max_by(self,
        key_selector: Callable[[TSource_co], Any],
        *args: Callable[[Any, Any], int],
    ) -> Any:
        if len(args) == 0:
            op = default_lt
        else:  # len(args) == 1
            comp = args[0]
            op = lambda l, r: comp(l, r) < 0
        return self._minmax_by_helper(key_selector, op)

    def min(self, *args: Callable[[TSource_co], Any]) -> Any:
        if len(args) == 0:
            result_selector: Any = identity
        else:  # len(args) == 1
            result_selector = args[0]
        return self._minmax_helper(
            result_selector,
            default_gt,
            self._raise_empty_sequence,
        )

    def min2(self, *args) -> Any:
        if len(args) == 1:
            result_selector, default = identity, args[0]
        else:  # len(args) == 2
            result_selector, default = args
        return self._minmax_helper(
            result_selector,
            default_gt,
            lambda: default,
        )

    def min_by(self,
        key_selector: Callable[[TSource_co], Any],
        *args: Callable[[Any, Any], int],
    ) -> Any:
        if len(args) == 0:
            op = default_gt
        else:  # len(args) == 1
            comp = args[0]
            op = lambda l, r: comp(l, r) > 0
        return self._minmax_by_helper(key_selector, op)

    def of_type(self, t_result: Type[TResult]) -> Enumerable[TResult]:
        return self.where(lambda e: isinstance(e, t_result)).cast(t_result)

    def order_by(self,
        key_selector: Callable[[TSource_co], TKey],
        *args: Callable[[TKey, TKey], int],
    ) -> OrderedEnumerable[TSource_co, TKey]:
        from .ordered_enumerable import OrderedEnumerable
        if len(args) == 1:
            comparer = args[0]
        else:  # len(args) == 2:
            comparer = None
        return OrderedEnumerable(
            self._get_iterable,
            None,
            key_selector,
            comparer,
            False,
        )

    def order_by_descending(self,
        key_selector: Callable[[TSource_co], TKey],
        *args: Callable[[TKey, TKey], int],
    ) -> OrderedEnumerable[TSource_co, TKey]:
        from .ordered_enumerable import OrderedEnumerable
        if len(args) == 1:
            comparer = args[0]
        else:  # len(args) == 2:
            comparer = None
        return OrderedEnumerable(
            self._get_iterable,
            None,
            key_selector,
            comparer,
            True,
        )

    def prepend(self, element: TSource_co) -> Enumerable[TSource_co]:  # type: ignore
        # see self.append()
        def inner():
            yield element
            yield from self
        return Enumerable(inner)

    @staticmethod
    def _raise_count_negative() -> NoReturn:
        raise InvalidOperationError('count must be nonnegative')

    @staticmethod
    def range(start: int, count: Optional[int]) -> Enumerable[int]:
        if count is not None:
            if count < 0:
                Enumerable._raise_count_negative()
            def inner(curr=start, cnt=count):  # type: ignore[misc]
                while cnt > 0:
                    yield curr
                    curr += 1
                    cnt -= 1
        else:
            def inner(curr=start):  # type: ignore[misc]
                while True:
                    yield curr
                    curr += 1
        return Enumerable(inner)

    @staticmethod
    def repeat(value: TResult, count: Optional[int] = None) -> Enumerable[TResult]:
        if count is not None:
            if count < 0:
                Enumerable._raise_count_negative()
            def inner(val=value, cnt=count):  # type: ignore[misc]
                while cnt > 0:
                    yield val
                    cnt -= 1
        else:
            def inner(val=value):  # type: ignore[misc]
                while True:
                    yield val
        return Enumerable(inner)

    def reverse(self) -> Enumerable[TSource_co]:
        return Enumerable(lambda: self._reversed_impl(fallback=True))

    def select(self, selector: Callable[[TSource_co], TResult]) -> Enumerable[TResult]:
        def inner():
            for elem in self:
                yield selector(elem)
        return Enumerable(inner)

    def select2(self, selector: Callable[[TSource_co, int], TResult]) -> Enumerable[TResult]:
        def inner():
            for i, elem in enumerate(self):
                yield selector(elem, i)
        return Enumerable(inner)

    def select_many(self,
        collection_selector: Callable[[TSource_co], Iterable[TCollection]],
        *args: Callable[[TSource_co, TCollection], TResult],
    ) -> Union[Enumerable[TCollection], Enumerable[TResult]]:
        if len(args) == 0:
            result_selector: Any = return_second
        else:  # len(args) == 1
            result_selector = args[0]
        def inner():
            for elem in self:
                for sub in collection_selector(elem):
                    yield result_selector(elem, sub)
        return Enumerable(inner)

    def select_many2(self,
        collection_selector: Callable[[TSource_co, int], Iterable[TCollection]],
        *args: Callable[[TSource_co, TCollection], TResult],
    ) -> Union[Enumerable[TCollection], Enumerable[TResult]]:
        if len(args) == 0:
            result_selector: Any = return_second
        else:  # len(args) == 1
            result_selector = args[0]
        def inner():
            for i, elem in enumerate(self):
                for sub in collection_selector(elem, i):
                    yield result_selector(elem, sub)
        return Enumerable(inner)

    def sequence_equal(self,
        second: Iterable[TSource_co],
        *args: Callable[..., bool],
    ) -> bool:
        if len(args) == 0:
            comparer = default_equal
        else:  # len(args) == 1
            comparer = args[0]

        me, she = iter(self), iter(second)
        while True:
            try:
                lhs = next(me)
            except StopIteration:
                try:
                    next(she)
                    return False
                except StopIteration:
                    return True
            try:
                rhs = next(she)
            except StopIteration:
                return False
            if not comparer(lhs, rhs):
                return False

    def _find_single(self, res):
        for elem in self:
            if res is not _signal:
                raise InvalidOperationError('Sequence does not contain exactly one element')
            res = elem
        return res

    def _find_single_with_predicate(self, res, predicate):
        for elem in self:
            if predicate(elem):
                if res is not _signal:
                    raise InvalidOperationError(
                        'There are multiple elements that satisfy condition: '
                        f'{res} vs. {elem}'
                    )
                res = elem
        return res

    def single(self, *args: Callable[[TSource_co], bool]) -> TSource_co:
        res: Any = _signal
        if len(args) == 0:
            res = self._find_single(res)
            if res is _signal:
                self._raise_empty_sequence()

        else:  # len(args) == 1
            predicate = args[0]
            res = self._find_single_with_predicate(res, predicate)
            if res is _signal:
                self._raise_no_such_element()
        return res

    def single2(self, *args):
        res: Any = _signal
        if len(args) == 1:
            default = args[0]
            res = self._find_single(res)

        else:  # len(args) == 2
            predicate, default = args
            res = self._find_single_with_predicate(res, predicate)
        if res is _signal:
            return default
        return res


    def skip(self, count: int) -> Enumerable[TSource_co]:
        def inner():
            iterator = iter(self)
            try:
                for _ in range(count):
                    next(iterator)
            except StopIteration:
                return
            yield from iterator
        return Enumerable(inner)

    def skip_last(self, count: int) -> Enumerable[TSource_co]:
        if count <= 0:
            return self.skip(0)
        def inner():
            iterator = iter(self)
            q = Deque()
            for elem in iterator:
                if len(q) == count:
                    while True:
                        yield q.pop()
                        q.appendleft(elem)
                        try:
                            elem = next(iterator)
                        except StopIteration:
                            break
                else:
                    q.appendleft(elem)
        return Enumerable(inner)

    def skip_while(self, predicate: Callable[[TSource_co], bool]) -> Enumerable[TSource_co]:
        def inner():
            iterator = iter(self)
            for elem in iterator:
                if not predicate(elem):
                    yield elem
                    break
            yield from iterator
        return Enumerable(inner)

    def skip_while2(self, predicate: Callable[[TSource_co, int], bool]) -> Enumerable[TSource_co]:
        def inner():
            iterator = iter(self)
            for i, elem in enumerate(iterator):
                if not predicate(elem, i):
                    yield elem
                    break
            yield from iterator
        return Enumerable(inner)

    def _sum_helper(self, selector, when_empty):
        iterator = iter(self)
        try:
            sum_ = selector(next(iterator))
        except StopIteration:
            return when_empty()

        for elem in iterator:
            sum_ += selector(elem)
        return sum_

    def sum(self, *args) -> Any:
        if len(args) == 0:
            selector: Any = identity
        else:  # len(args) == 1
            selector = args[0]
        return self._sum_helper(selector, lambda: 0)

    def sum2(self, *args) -> Any:
        if len(args) == 1:
            selector, default = identity, args[0]
        else: # len(args) == 2
            selector, default = args
        return self._sum_helper(selector, lambda: default)

    def take(self, count: Union[int, slice]) -> Enumerable[TSource_co]:
        if isinstance(count, int):
            def inner(c: int = count):
                iterator = iter(self)
                try:
                    for _ in range(c):
                        yield next(iterator)
                except StopIteration:
                    return
            return Enumerable(inner)
        else:  # isinstance(count, slice)
            return self.elements_in(count)

    def take_last(self, count: int) -> Enumerable[TSource_co]:
        if count <= 0:
            return self.empty()
        def inner():
            iterator = iter(self)
            try:
                q = Deque((next(iterator),))
            except StopIteration:
                return
            for elem in iterator:
                if len(q) == count:
                    while True:
                        q.popleft()
                        q.append(elem)
                        try:
                            elem = next(iterator)
                        except StopIteration:
                            break
                else:
                    q.append(elem)
            yield from q
        return Enumerable(inner)

    def take_while(self, predicate: Callable[[TSource_co], bool]) -> Enumerable[TSource_co]:
        def inner():
            for elem in self:
                if not predicate(elem):
                    break
                yield elem
        return Enumerable(inner)

    def take_while2(self, predicate: Callable[[TSource_co, int], bool]) -> Enumerable[TSource_co]:
        def inner():
            for i, elem in enumerate(self):
                if not predicate(elem, i):
                    break
                yield elem
        return Enumerable(inner)

    def to_dict(self,
        key_selector: Callable[[TSource_co], TKey],
        *args: Callable[[TSource_co], TValue],
    ) -> Union[Dict[TKey, TValue], Dict[TKey, TSource_co]]:
        if len(args) == 0:
            value_selector: Any = identity
        else:  # len(args) == 1
            value_selector = args[0]
        return {key_selector(e): value_selector(e) for e in self}

    def to_set(self) -> Set[TSource_co]:
        return {e for e in self}

    def to_list(self) -> List[TSource_co]:
        return [e for e in self]

    def to_lookup(self,
        key_selector: Callable[[TSource_co], TKey],
        *args: Callable[[TSource_co], TValue],
    ) -> Union[Lookup[TKey, TValue], Lookup[TKey, TSource_co]]:
        from .lookup import Lookup
        if len(args) == 0:
            value_selector: Any = identity
        else:  # len(args) == 1
            value_selector = args[0]
        res = Lookup(self, key_selector, value_selector)
        return res

    def union(self, second: Iterable[TSource_co]) -> Enumerable[TSource_co]:
        # TODO: optimise chained .union() calls
        return self.union_by(second, identity)

    def union_by(self,
        second: Iterable[TSource_co],
        key_selector: Callable[[TSource_co], object],
    ) -> Enumerable[TSource_co]:
        def inner():
            s = ComposeSet()
            for elem in self.concat(second):
                key = key_selector(elem)
                if key in s:
                    continue
                s.add(key)
                yield elem
        return Enumerable(inner)

    def where(self, predicate: Callable[[TSource_co], bool]) -> Enumerable[TSource_co]:
        def inner():
            for elem in self:
                if predicate(elem):
                    yield elem
        return Enumerable(inner)

    def where2(self, predicate: Callable[[TSource_co, int], bool]) -> Enumerable[TSource_co]:
        def inner():
            for i, elem in enumerate(self):
                if predicate(elem, i):
                    yield elem
        return Enumerable(inner)

    def zip(self, *iters: Iterable[Any]) -> Enumerable[Any]:
        def inner():
            yield from zip(self, *iters)
        return Enumerable(inner)

    def zip2(self, *iters_and_result_selector: Any) -> Enumerable[Any]:
        iters = iters_and_result_selector[:-1]
        result_selector = iters_and_result_selector[-1]
        def inner():
            for tup in zip(self, *iters):
                yield result_selector(*tup)
        return Enumerable(inner)

    def elements_in(self, *args) -> Enumerable[TSource_co]:
        if len(args) == 1:
            index = args[0]
            return self._getitem_impl(index, fallback=True)  # type: ignore
        elif len(args) == 2:
            start, stop = args
            return self.elements_in(start, stop, 1)
        else:  # len(args) == 3
            return self.elements_in(slice(*args))

    def to_tuple(self) -> Tuple[TSource_co, ...]:
        return tuple(e for e in self)



class TypesLinqError(Exception):
    '''
    ```py
    from types_linq import TypesLinqError
    ```

    Types-linq has run into problems.
    '''


class InvalidOperationError(TypesLinqError, ValueError):
    '''
    ```py
    from types_linq import InvalidOperationError
    ```

    Exception raised when a call is invalid for the object's current state.
    '''


class IndexOutOfRangeError(TypesLinqError, IndexError):
    '''
    ```py
    from types_linq import IndexOutOfRangeError
    ```

    An `IndexError` with types-linq flavour.
    '''



from __future__ import annotations

from typing import Any, Callable, Deque, Iterable, Iterator, List, Optional, TYPE_CHECKING, Tuple

if TYPE_CHECKING:
    from .extrema_enumerable import ExtremaEnumerable

from .more_enums import RankMethods
from .more_error import DirectedGraphNotAcyclicError
from ..enumerable import Enumerable
from ..util import (
    ComposeMap,
    ComposeSet,
    default_equal,
    identity,
)
from ..more_typing import (
    TAccumulate,
    TKey,
    TSource,
    TSource_co,
)


class MoreEnumerable(Enumerable[TSource_co]):

    def aggregate_right(self, *args) -> Any:
        # NOTE: a copy of the sequence is made in this call because it falls back
        it = iter(self.reverse())
        if len(args) == 3:
            seed, func, result_selector = args
        elif len(args) == 2:
            seed, func, result_selector = args[0], args[1], identity
        else:  # len(args) == 1
            func, result_selector = args[0], identity
            try:
                seed = next(it)
            except StopIteration:
                self._raise_empty_sequence()

        for elem in it:
            seed = func(elem, seed)
        return result_selector(seed)

    def as_more(self) -> MoreEnumerable[TSource_co]:  # pyright: ignore[reportIncompatibleMethodOverride]
        return self

    def consume(self) -> None:
        for _ in self:
            ...

    def cycle(self, count: Optional[int] = None) -> MoreEnumerable[TSource_co]:
        if count is not None:
            if count < 0:
                self._raise_count_negative()  # type: ignore
            elif count == 0:
                return MoreEnumerable(())
            elif count == 1:
                return self

        def inner(cnt=count):
            memo: List[TSource_co] = []
            for elem in self:
                memo.append(elem)
                yield elem
            while cnt is None or cnt > 1:
                yield from memo
                if cnt is not None:
                    cnt -= 1
        return MoreEnumerable(inner)

    def enumerate(self, start: int = 0) -> MoreEnumerable[Tuple[int, TSource_co]]:
        return MoreEnumerable(lambda: enumerate(self, start))

    def except_by2(self,
        second: Iterable[TSource_co],
        key_selector: Callable[[TSource_co], object],
    ) -> MoreEnumerable[TSource_co]:
        def inner():
            s = ComposeSet(key_selector(s) for s in second)
            for elem in self:
                key = key_selector(elem)
                if key in s:
                    continue
                s.add(key)
                yield elem
        return MoreEnumerable(inner)

    def flatten(self, *args: Callable[[Iterable[Any]], bool]) -> MoreEnumerable[Any]:
        if len(args) == 0:
            return self.flatten(lambda x: not isinstance(x, (str, bytes)))
        else:  # len(args) == 1
            predicate = args[0]
            return self.flatten2(lambda x: x
                if isinstance(x, Iterable) and predicate(x)
                else None)

    def flatten2(self, selector: Callable[[Any], Optional[Iterable[object]]]) -> MoreEnumerable[Any]:
        def inner():
            stack: List[Iterator[object]] = [iter(self)]
            while stack:
                it = stack.pop()
                while True:
                    try:
                        elem = next(it)
                    except StopIteration:
                        break
                    nested = selector(elem)
                    if nested is not None:
                        stack.append(it)
                        it = iter(nested)
                        continue
                    yield elem
        return MoreEnumerable(inner)

    def for_each(self, action: Callable[[TSource_co], object]) -> None:
        for elem in self:
            action(elem)

    def for_each2(self, action: Callable[[TSource_co, int], object]) -> None:
        for i, elem in enumerate(self):
            action(elem, i)

    def interleave(self, *iters: Iterable[TSource_co]) -> MoreEnumerable[TSource_co]:
        def inner():
            its = [iter(self)]
            for iter_ in iters:
                its.append(iter(iter_))
            while its:
                i = 0
                while i < len(its):
                    try:
                        yield next(its[i])
                        i += 1
                    except StopIteration:
                        its.pop(i)
        return MoreEnumerable(inner)

    def maxima_by(self,
        selector: Callable[[TSource_co], TKey],
        *args: Callable[[TKey, TKey], int],
    ) -> ExtremaEnumerable[TSource_co, TKey]:
        from .extrema_enumerable import ExtremaEnumerable
        if len(args) == 0:
            comparer = None
        else:  # len(args) == 1
            comparer = args[0]
        return ExtremaEnumerable(self, selector, comparer, False)

    def minima_by(self,
        selector: Callable[[TSource_co], TKey],
        *args: Callable[[TKey, TKey], int],
    ) -> ExtremaEnumerable[TSource_co, TKey]:
        from .extrema_enumerable import ExtremaEnumerable
        if len(args) == 0:
            comparer = None
        else:  # len(args) == 1
            comparer = args[0]
        return ExtremaEnumerable(self, selector, comparer, True)

    def pipe(self, action: Callable[[TSource_co], object]) -> MoreEnumerable[TSource_co]:
        def inner():
            for elem in self:
                action(elem)
                yield elem
        return MoreEnumerable(inner)

    def pre_scan(self,
        identity: TAccumulate,
        transformation: Callable[[TAccumulate, TSource_co], TAccumulate],
    ) -> MoreEnumerable[TAccumulate]:
        def inner(aggregator: TAccumulate = identity):
            it = iter(self)
            try:
                past = next(it)
            except StopIteration:
                return
            yield identity
            for elem in it:
                aggregator = transformation(aggregator, past)
                yield aggregator
                past = elem
        return MoreEnumerable(inner)

    def rank(self,
        *args: Callable[[TSource_co, TSource_co], int],
        method: RankMethods = RankMethods.dense,
    ) -> MoreEnumerable[int]:
        return self.rank_by(identity, *args, method=method)

    def rank_by(self,
        key_selector: Callable[[TSource_co], TKey],
        *args: Callable[[TKey, TKey], int],
        method: RankMethods = RankMethods.dense,
     ) -> MoreEnumerable[int]:
        if len(args) == 0:
            # it is sufficient to have only equality
            comparer = lambda l, r: 0 if l == r else 1
        else:  # len(args) == 1
            comparer = args[0]

        def inner():
            # avoid enumerating twice
            copy = self.select(key_selector).to_list()
            if not copy:
                return
            l = len(copy)
            ordered = MoreEnumerable(range(l)) \
                .order_by_descending(lambda i: copy[i], *args).to_list()
            rank_map = [0] * l
            rank_map[ordered[0]] = 1
            rank = 1
            consec = 1
            for i in range(1, l):
                if method == RankMethods.ordinal:
                    rank += 1
                elif comparer(copy[ordered[i - 1]], copy[ordered[i]]) != 0:
                    rank += consec
                    consec = 1
                elif method == RankMethods.competitive:
                    consec += 1
                rank_map[ordered[i]] = rank
            del ordered
            for i in rank_map:
                yield i
        return MoreEnumerable(inner)

    def run_length_encode(self,
        *args: Callable[[TSource_co, TSource_co], bool],
    ) -> MoreEnumerable[Tuple[TSource_co, int]]:
        if len(args) == 0:
            comparer = default_equal
        else:
            comparer = args[0]
        def inner():
            iterator = iter(self)
            try:
                prev_elem = next(iterator)
            except StopIteration:
                return
            count = 1
            while True:
                try:
                    elem = next(iterator)
                except StopIteration:
                    break
                if comparer(prev_elem, elem):
                    count += 1
                else:
                    yield (prev_elem, count)
                    prev_elem = elem
                    count = 1
            yield (prev_elem, count)
        return MoreEnumerable(inner)

    def scan(self, *args) -> Any:
        if len(args) == 2:
            seed, transformation = args
        else:  # len(args) == 1
            transformation = args[0]
        def inner():
            it = iter(self)
            nonlocal seed
            if len(args) == 1:
                try:
                    seed = next(it)
                except StopIteration:
                    return
            yield seed
            for elem in it:
                seed = transformation(seed, elem)
                yield seed
        return MoreEnumerable(inner)

    def scan_right(self, *args) -> Any:
        if len(args) == 2:
            seed, func = args
        else:  # len(args) == 1
            func = args[0]
        def inner():
            # NOTE: a copy of the sequence is made in this call because it falls back
            it = iter(self.reverse())
            nonlocal seed
            if len(args) == 1:
                try:
                    seed = next(it)
                except StopIteration:
                    return
            results = []
            results.append(seed)
            for elem in it:
                seed = func(elem, seed)
                results.append(seed)
            yield from reversed(results)
        return MoreEnumerable(inner)

    def segment(self,
        new_segment_predicate: Callable[[TSource_co], bool],
    ) -> MoreEnumerable[MoreEnumerable[TSource_co]]:
        return self.segment3(lambda x, p, i: new_segment_predicate(x))

    def segment2(self,
        new_segment_predicate: Callable[[TSource_co, int], bool],
    ) -> MoreEnumerable[MoreEnumerable[TSource_co]]:
        return self.segment3(lambda x, _, i: new_segment_predicate(x, i))

    def segment3(self,
        new_segment_predicate: Callable[[TSource_co, TSource_co, int], bool],
    ) -> MoreEnumerable[MoreEnumerable[TSource_co]]:
        def inner():
            it = iter(self)
            try:
                prev = next(it)
            except StopIteration:
                return
            segment = [prev]
            for i, elem in enumerate(it, 1):
                if new_segment_predicate(elem, prev, i):
                    yield MoreEnumerable(segment)
                    segment = [elem]
                else:
                    segment.append(elem)
                prev = elem
            yield MoreEnumerable(segment)
        return MoreEnumerable(inner)

    @staticmethod
    def traverse_breath_first(
        root: TSource,
        children_selector: Callable[[TSource], Iterable[TSource]],
    ) -> MoreEnumerable[TSource]:
        def inner():
            queue = Deque((root,))
            while queue:
                elem = queue.popleft()
                yield elem
                for child in children_selector(elem):
                    queue.append(child)
        return MoreEnumerable(inner)

    @staticmethod
    def traverse_depth_first(
        root: TSource,
        children_selector: Callable[[TSource], Iterable[TSource]],
    ) -> MoreEnumerable[TSource]:
        def inner():
            stack = [root]
            while stack:
                elem = stack.pop()
                yield elem
                children = [*children_selector(elem)]
                for child in reversed(children):
                    stack.append(child)
        return MoreEnumerable(inner)

    def traverse_topological(self,
        children_selector: Callable[[TSource_co], Iterable[TSource_co]],
    ) -> MoreEnumerable[TSource_co]:
        return self.traverse_topological2(children_selector, identity)

    def traverse_topological2(self,
        children_selector: Callable[[TSource_co], Iterable[TSource_co]],
        key_selector: Callable[[TSource_co], object],
    ) -> MoreEnumerable[TSource_co]:
        def inner():
            stack: List[Tuple[TSource_co, bool]] = []
            visited = ComposeMap()
            result: List[TSource_co] = []  # post order
            result_keys = ComposeSet()
            # NOTE: a copy of the sequence is made in this call because it falls back
            for root in self.reverse():
                if not visited.get(key_selector(root)):
                    stack.append((root, False))
                while stack:
                    node, done = stack.pop()
                    if done:
                        # detect cycle. normally edges from node will go to nodes in the
                        # result list. otherwise we found cycle
                        for child in children_selector(node):
                            if key_selector(child) not in result_keys:
                                raise DirectedGraphNotAcyclicError((node, child))
                        # good
                        result.append(node)
                        result_keys.add(key_selector(node))
                        continue
                    visited[key_selector(node)] = True
                    stack.append((node, True))
                    for child in children_selector(node):
                        if not visited.get(key_selector(child)):
                            stack.append((child, False))
            while result:
                yield result.pop()
        return MoreEnumerable(inner)



from __future__ import annotations
from typing import Any, Callable, Generic, Iterable, Iterator, Optional

from .enumerable import Enumerable

from .more_typing import (
    TSource_co,
    TKey,
    TKey2,
)


class OrderedEnumerable(Enumerable[TSource_co], Generic[TSource_co, TKey]):

    _parent: Optional[OrderedEnumerable[TSource_co, Any]]
    _key_selector: Callable[[TSource_co], TKey]
    _comparer: Optional[Callable[[TKey, TKey], int]]
    _descending: bool

    def __init__(self,
        source: Callable[[], Iterable[TSource_co]],  # pass along
        parent: Optional[OrderedEnumerable[TSource_co, Any]],
        key_selector: Callable[[TSource_co], TKey],
        comparer: Optional[Callable[[TKey, TKey], int]],
        descending: bool,
    ):
        super().__init__(source)
        self._parent = parent
        self._key_selector = key_selector
        self._comparer = comparer
        self._descending = descending

    def _get_iterable(self) -> Iterator[TSource_co]:
        lst = [elem for elem in super()._get_iterable()]
        curr = self
        while curr is not None:
            comparer = curr._comparer
            if comparer is None:
                # comparer-less overload for order_by(), etc. ensures TSource_co must support __lt__()
                key = curr._key_selector  # type: ignore
            else:
                selector = curr._key_selector
                class key:
                    def __init__(self, elem):
                        self.elem = elem
                    def __lt__(self, __o: key):
                        return comparer(selector(self.elem), selector(__o.elem)) < 0  # type: ignore
                    # it is OK to omit __eq__() because python sort only uses __lt__()
            lst.sort(key=key, reverse=curr._descending)
            curr = curr._parent
        yield from lst

    def create_ordered_enumerable(self,
        key_selector: Callable[[TSource_co], TKey2],
        comparer: Optional[Callable[[TKey2, TKey2], int]],
        descending: bool,
    ) -> OrderedEnumerable[TSource_co, TKey2]:
        return OrderedEnumerable(
            self._get_iterable,
            self,
            key_selector,
            comparer,
            descending,
        )

    def then_by(self,
        key_selector: Callable[[TSource_co], TKey2],
        *args: Callable[[TKey2, TKey2], int],
    ) -> OrderedEnumerable[TSource_co, TKey2]:
        if len(args) == 1:
            comparer = args[0]
        else:  # len(args) == 2:
            comparer = None
        return self.create_ordered_enumerable(
            key_selector,
            comparer,
            False,
        )

    def then_by_descending(self,
        key_selector: Callable[[TSource_co], TKey2],
        *args: Callable[[TKey2, TKey2], int],
    ) -> OrderedEnumerable[TSource_co, TKey2]:
        if len(args) == 1:
            comparer = args[0]
        else:  # len(args) == 2:
            comparer = None
        return self.create_ordered_enumerable(
            key_selector,
            comparer,
            True,
        )



from .enumerable import Enumerable
from .types_linq_error import TypesLinqError, InvalidOperationError, IndexOutOfRangeError


__all__ = [
    'Enumerable',
    'TypesLinqError',
    'InvalidOperationError',
    'IndexOutOfRangeError',
]



# class 'MoreError' is nonexistent
from __future__ import annotations
from typing import Tuple

from ..types_linq_error import InvalidOperationError


class DirectedGraphNotAcyclicError(InvalidOperationError):
    '''
    ```py
    from types_linq.more import DirectedGraphNotAcyclicError
    ```

    Exception raised when a cycle exists in a graph.

    Revisions
        ~ v1.2.1: New.
    '''

    def __init__(self, cycle: Tuple[object, object]) -> None:
        super().__init__('cycle detected')
        self._cycle = cycle

    @property
    def cycle(self) -> Tuple[object, object]:
        '''
        The two elements (A, B) in this tuple are part of a cycle. There exists an edge from A to B,
        and a path from B back to A. A and B may be identical.

        Example
        ```py
        >>> adj = { 5: [2, 0], 4: [0, 1], 2: [3], 3: [1, 5] }
        >>> try:
        >>>     MoreEnumerable([5, 4]).traverse_topological(lambda x: adj.get(x, [])) \\
        >>>         .consume()
        >>> except DirectedGraphNotAcyclicError as e:
        >>>     print(e.cycle)
        (3, 5)  # 3 -> 5 -> 2 -> 3
        ```
        '''
        return self._cycle



from .more_enumerable import MoreEnumerable
from .more_enums import RankMethods
from .more_error import DirectedGraphNotAcyclicError


__all__ = [
    'MoreEnumerable',
    'RankMethods',
    'DirectedGraphNotAcyclicError',
]



from __future__ import annotations
from typing import Callable, Iterable

from .enumerable import Enumerable
from .grouping import Grouping
from .util import ComposeMap

from .more_typing import (
    TSource,
    TKey_co,
    TValue_co,
    TResult,
)


# TODO: Wish to support Mapping[TKey, TValue], but its default mixin methods are doing something
# weird.
class Lookup(Enumerable[Grouping[TKey_co, TValue_co]]):
    '''
    ```py
    from types_linq.lookup import Lookup
    ```

    A lookup is a one-to-many dictionary. It maps keys to Enumerable sequences of values.

    Users should not construct instances of this class directly. Use `Enumerable.to_lookup()`
    instead.
    '''

    _groupings: ComposeMap[TKey_co, Grouping[TKey_co, TValue_co]]

    def __init__(self,
        source: Iterable[TSource],
        key_selector: Callable[[TSource], TKey_co],
        value_selector: Callable[[TSource], TValue_co],
    ):
        self._groupings = ComposeMap()

        for src in source:
            key = key_selector(src)
            elem = value_selector(src)
            if key not in self._groupings:
                self._groupings[key] = Grouping(key)
            self._groupings[key]._append(elem)

        super().__init__(self._groupings.values())

    def __contains__(self, value: object) -> bool:
        '''
        Tests whether key is in the lookup.
        '''
        return value in self._groupings

    def __len__(self) -> int:
        '''
        Gets the number of key-collection pairs.
        '''
        return len(self._groupings)

    def __getitem__(self, key: TKey_co) -> Enumerable[TValue_co]:  # type: ignore
        '''
        Gets the collection of values indexed by the specified key, or empty if no such key
        exists.
        '''
        if key in self._groupings:
            return self._groupings[key]
        return Enumerable.empty()  # type: ignore

    def apply_result_selector(self,
        result_selector: Callable[[TKey_co, Enumerable[TValue_co]], TResult],
    ) -> Enumerable[TResult]:
        '''
        Applies a transform function to each key and its associated values, then returns the
        results.
        '''
        def inner():
            for key, grouping in self._groupings.items():
                yield result_selector(key, grouping)
        return Enumerable(inner)

    def contains(self, value: object) -> bool:  # type: ignore[override]
        '''
        Tests whether key is in the lookup.
        '''
        return value in self

    @property
    def count(self) -> int:    # type: ignore[override]
        '''
        Gets the number of key-collection pairs.
        '''
        return len(self)



from enum import Enum, auto


class RankMethods(Enum):
    '''
    ```py
    from types_linq.more import RankMethods
    ```

    Enumeration to select different methods of assigning rankings when breaking
    [ties](https://en.wikipedia.org/wiki/Ranking#Strategies_for_assigning_rankings).

    Revisions
        ~ v1.2.1: New.
    '''
    dense = auto()
    '''
    Items that compare equally receive the same ranking, and the next items get the immediately
    following ranking. *(1223)*
    '''

    competitive = auto()
    '''
    Items that compare equally receive the same highest ranking, and gaps are left out. *(1224)*
    '''

    ordinal = auto()
    '''
    Each item receives unique rankings. *(1234)*
    '''



from setuptools import setup

with open('README.md', 'r', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='types-linq',
    version='v1.2.1',
    url='https://github.com/cleoold/types-linq',
    license='BSD 2-Clause License',
    author='cleoold',
    description='Standard sequence helper methods with full typing support',
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=['types_linq', 'types_linq.more'],
    package_data={
        '': ['*.pyi', 'py.typed'],
    },
    zip_safe=False,
    python_requires='>=3.7',
    extras_require={':python_version<"3.8"': ['typing_extensions']},
    platforms='any',
    classifiers=[
        'Topic :: Utilities',
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Typing :: Typed',
    ],
)
